#!/usr/bin/env python3
import os, os.path, sys, re
from .argparse_compat import argparse
from .custom_arguments import (
    add_libname_arguments,
    update_env_with_libnames,
    add_logging_arguments,
    process_logging_arguments,
    get_parser_name_mapping,
    LOG_ALWAYS,
    DEFAULT_VERBOSITY,
)
from .split_file import UnsupportedCoqVersionError
from .import_util import (
    get_file_statements_insert_references,
    sort_files_by_dependency,
    classify_require_kind,
    EXPORT,
    REQUIRE_EXPORT,
)
from .file_util import write_to_file
from .memoize import memoize
from .minimizer_drivers import run_binary_search
from . import diagnose_error

__all__ = ["main"]

parser = argparse.ArgumentParser(description="Remove useless Requires in a file")
parser.add_argument("input_files", metavar="INFILE", nargs="*", type=argparse.FileType("r"), help=".v files to update")
parser.add_argument(
    "--in-place",
    "-i",
    metavar="SUFFIX",
    dest="suffix",
    nargs="?",
    type=str,
    default="",
    help="update files in place (makes backup if SUFFIX supplied)",
)
parser.add_argument(
    "--update-all",
    "--all",
    dest="update_all",
    action="store_const",
    default=False,
    const=True,
    help=("also update all .v files listed in any _CoqProject file passed to -f (implies --in-place, requires -f)"),
)
parser.add_argument(
    "--absolutize",
    dest="absolutize",
    action="store_const",
    default=False,
    const=True,
    help=("Replace Requires with fully qualified versions."),
)
parser.add_argument(
    "--timeout",
    dest="timeout",
    metavar="SECONDS",
    type=int,
    default=-1,
    help=(
        "Use a timeout; make sure Coq is "
        + "killed after running for this many seconds. "
        + "If 0, there is no timeout.  If negative, then "
        + "twice the initial run of the script is used.\n\n"
        + "Default: -1"
    ),
)
parser.add_argument(
    "--no-keep-exports",
    dest="keep_exports",
    action="store_const",
    default=True,
    const=False,
    help=("Allow the removal of Require lines that have Export in them"),
)
parser.add_argument("--no-timeout", dest="timeout", action="store_const", const=0, help=("Do not use a timeout"))
parser.add_argument(
    "--keep-going",
    "-k",
    dest="keep_going",
    action="store_const",
    const=True,
    default=False,
    help=("Keep going when some files can't be minimized."),
)
parser.add_argument(
    "--coqbin",
    metavar="COQBIN",
    dest="coqbin",
    type=str,
    default="",
    help="The path to a folder containing the coqc and coqtop programs.",
)
parser.add_argument(
    "--coqc", metavar="COQC", dest="coqc", type=str, default=None, action="append", help="The path to the coqc program."
)
add_libname_arguments(parser)
add_logging_arguments(parser)


def mark_exports(state, keep_exports):
    return tuple(
        (
            text_as_bytes,
            ranges,
            bool(keep_exports and ranges and classify_require_kind(text_as_bytes, ranges) in (EXPORT, REQUIRE_EXPORT)),
        )
        for text_as_bytes, ranges in state
    )


SKIP = "SKIP"
REMOVE = "REMOVE"


def trailing_whitespace(text):
    return "".join(sorted(re.search(r"(\s*)$", text, flags=re.DOTALL).groups()[0]))


def leading_whitespace(text):
    return "".join(sorted(re.search(r"^(\s*)", text, flags=re.DOTALL).groups()[0]))


def whitespace_to_key(ws):
    return (ws.count("\n"), ws.count("\t"), ws.count(" "), len(ws), ws)


def gobble_whitespace_text(before, after):
    before_ws, after_ws = trailing_whitespace(before), leading_whitespace(after)
    assert whitespace_to_key("") < whitespace_to_key(" ")  # sanity check
    if before_ws and after_ws:
        if whitespace_to_key(before_ws) < whitespace_to_key(after_ws):
            return (before[: -len(before_ws)], after)
        else:
            return (before, after[len(after_ws) :])
    elif before_ws:
        return (before[: -len(before_ws)], after)
    elif after_ws:
        return (before, after[len(after_ws) :])
    else:
        return (before, after)


def gobble_whitespace(before, after):
    assert before is bytes(before)
    assert after is bytes(after)
    before, after = gobble_whitespace_text(before.decode("utf-8"), after.decode("utf-8"))
    return (before.encode("utf-8"), after.encode("utf-8"))


def step_state(state, action):
    ret = []
    state = list(state)
    while len(state) > 0:
        (text_as_bytes, references, force_keep), state = state[0], state[1:]
        if references:
            (start, end, _loc, _append, _ty), new_references = references[0], tuple(references[1:])
            if action == SKIP or action is None:
                ret.append((text_as_bytes, new_references, force_keep))
            elif action == REMOVE and not force_keep:
                if new_references:  # still other imports, safe to just remove
                    pre_text_as_bytes, post_text_as_bytes = gobble_whitespace(
                        text_as_bytes[:start], text_as_bytes[end:]
                    )
                    ret.append((pre_text_as_bytes + post_text_as_bytes, new_references, force_keep))
                else:  # no other imports; remove this line completely
                    if ret and state:
                        prev_text_as_bytes, post_text_as_bytes = gobble_whitespace(ret[-1][0], state[0][0])
                        ret[-1] = (prev_text_as_bytes,) + ret[-1][1:]
                        state[0] = (post_text_as_bytes,) + state[0][1:]
                    elif ret:
                        ret[-1] = (ret[-1][0].rstrip(),) + ret[-1][1:]
                    elif state:
                        state[0] = (state[0][0].lstrip(),) + state[0][1:]
            else:
                raise ValueError
            ret.extend(state)
            return tuple(ret)
        else:
            ret.append((text_as_bytes, references, force_keep))
    return None


def state_to_contents(state):
    return "".join(reversed([v[0].decode("utf-8") for v in state]))


# the higher verbose_base, the less verbose we are, since we start at a higher level by default
def make_check_state(original_contents, verbose_base=4 - DEFAULT_VERBOSITY, **kwargs):
    # original_contents is str
    expected_output, orig_cmds, orig_retcode, runtime = diagnose_error.get_coq_output(
        kwargs["coqc"], kwargs["coqc_args"], original_contents, kwargs["timeout"], verbose_base=2, **kwargs
    )

    @memoize
    def check_contents(contents):
        output, cmds, retcode, runtime = diagnose_error.get_coq_output(
            kwargs["coqc"], kwargs["coqc_args"], contents, kwargs["timeout"], verbose_base=2, **kwargs
        )
        # TODO: Should we be checking the error message and the retcode and the output, or just the retcode?
        retval = diagnose_error.has_error(output) or output != expected_output or retcode != orig_retcode
        if retval:
            kwargs["log"](
                'Failed change.  Error when running "%s":\n%s' % ('" "'.join(cmds), output), level=verbose_base - 1
            )
        else:
            kwargs["log"]("Successful change", level=verbose_base)
            kwargs["log"]('New contents:\n"""\n%s\n"""' % contents, level=verbose_base + 1)
        return not retval

    def check_state(state):
        return check_contents(state_to_contents(state))

    return check_state


def make_save_state(filename, **kwargs):
    def save_state(state, final=False):
        contents = state_to_contents(state)
        if kwargs["inplace"]:
            do_backup = kwargs["suffix"] is not None and len(kwargs["suffix"]) > 0
            write_to_file(filename, contents, do_backup=do_backup, backup_ext=kwargs["suffix"])
        elif final:
            print(contents)

    return save_state


def main():
    args = process_logging_arguments(parser.parse_args())

    def prepend_coqbin(prog):
        if isinstance(prog, str):
            prog = (prog,)
        if args.coqbin != "":
            return (os.path.join(args.coqbin, prog[0]), *prog[1:])
        else:
            return tuple(prog)

    env = {
        "log": args.log,
        "keep_exports": args.keep_exports,
        "keep_going": args.keep_going,
        "coqc": prepend_coqbin(args.coqc or "coqc"),
        "coqc_args": (args.coq_args if args.coq_args else tuple()),
        "timeout": args.timeout,
        "inplace": args.suffix != "",  # it's None if they passed no argument, and '' if they didn't pass -i
        "suffix": args.suffix,
        "input_files": tuple(f.name for f in args.input_files),
        "cli_mapping": get_parser_name_mapping(parser),
    }
    update_env_with_libnames(env, args)

    for f in args.input_files:
        f.close()

    if args.update_all:
        if env["_CoqProject"] is None:
            parser.error("--update-all given without -f")
            sys.exit(1)
        else:
            env["inplace"] = True
            if env["suffix"] == "":
                env["suffix"] = None
            env["input_files"] = tuple(list(env["input_files"]) + list(env["_CoqProject_v_files"]))
        if len(env["input_files"]) == 0:
            parser.error("no .v files listed in %s" % args.CoqProjectFile)
    elif len(env["input_files"]) == 0:
        parser.error("not enough arguments (-f COQPROJECTFILE with .v files is required if no .v files are given)")

    for dirname, libname in env["libnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-R", dirname, libname])
    for dirname, libname in env["non_recursive_libnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-Q", dirname, libname])
    for dirname in env["ocaml_dirnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-I", dirname])

    env["input_files"] = sort_files_by_dependency(env["input_files"], update_globs=True, **env)

    try:
        failed = []
        for name in env["input_files"]:
            try:
                annotated_contents = get_file_statements_insert_references(
                    name,
                    absolutize=(("lib",) if args.absolutize else tuple()),
                    update_globs=True,
                    types=("lib",),
                    appends=("<>",),
                    **env,
                )
                if annotated_contents is None:
                    env["log"]("ERROR: Failed to get references for %s" % name, level=LOG_ALWAYS)
                    failed.append((name, "failed to get references"))
                    if env["keep_going"]:
                        continue
                    else:
                        break
                annotated_contents = mark_exports(tuple(reversed(annotated_contents)), env["keep_exports"])
                save_state = make_save_state(name, **env)
                check_state = make_check_state(state_to_contents(annotated_contents), **env)
                verbose_check_state = make_check_state(
                    state_to_contents(annotated_contents), verbose_base=DEFAULT_VERBOSITY, **env
                )
                env["log"]("Running coq on initial contents...")
                if not verbose_check_state(annotated_contents):
                    env["log"]("ERROR: Failed to update %s" % name, level=LOG_ALWAYS)
                    failed.append((name, "failed to update"))
                    if env["keep_going"]:
                        continue
                    else:
                        break
                valid_actions = (REMOVE,)
                final_state = run_binary_search(annotated_contents, check_state, step_state, save_state, valid_actions)
                if final_state is not None:
                    if not check_state(final_state):
                        env["log"]("Internal error: Inconsistent final state on %s..." % name, level=LOG_ALWAYS)
                        failed.append((name, "inconsistent final state"))
                        if not env["keep_going"]:
                            break
                    else:
                        env["log"]("Saving final version of %s..." % name)
                        save_state(final_state, final=True)
            except KeyboardInterrupt:
                raise
            except SystemExit:
                raise
            except BaseException as e:
                if env["keep_going"]:
                    env["log"]("Failure on %s with error %s" % (name, repr(e)), level=LOG_ALWAYS)
                    failed.append((name, e))
                else:
                    raise e
        if failed:
            env["log"]("The following files failed:", level=LOG_ALWAYS)
            for name, e in failed:
                env["log"](name, level=LOG_ALWAYS)
            sys.exit(1)
    except UnsupportedCoqVersionError:
        env["log"]("ERROR: Your version of coqc (%s) does not support -time" % env["coqc"], level=LOG_ALWAYS)
        sys.exit(1)


if __name__ == "__main__":
    main()



import os, sys, re
from difflib import SequenceMatcher
from typing import List
from .argparse_compat import argparse

__all__ = [
    "prompt",
    "yes_no_prompt",
    "b",
    "s",
    "cmp_compat",
    "PY3",
    "BooleanOptionalAction",
    "argparse",
    "raw_input",
    "re_escape",
    "slice_string_at_bytes",
    "len_in_bytes",
    "shlex_quote",
    "shlex_join",
    "resource_path",
    "group_by",
]

BooleanOptionalAction = argparse.BooleanOptionalAction

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

if sys.version_info < (3,):
    PY3 = False

    def b(x):
        return x

    def s(x):
        return x

else:
    PY3 = True

    def b(x) -> bytes:
        if x is not None:
            return x.encode()
        return x

    def s(x) -> str:
        # Sometimes we get str rather than bytes??? cf https://gitlab.com/coq/coq/-/jobs/544269051
        if hasattr(x, "decode"):
            return x.decode("utf-8", "ignore")
        return x

    def cmp(x, y):
        if x < y:
            return -1
        if y < x:
            return 1
        return 0

    raw_input = input

if sys.version_info < (3, 7):
    # see https://bugs.python.org/issue29995
    def re_escape(*args, **kwargs):
        ret = re.escape(*args, **kwargs)
        for ch in ':"':
            ret = ret.replace("\\" + ch, ch)
        return ret

else:
    re_escape = re.escape

cmp_compat = cmp


def prompt(
    options,
    case_sensitive=False,
    strip=True,
    sep="/",
    prefix="Please enter ",
    postfix=": ",
    yes_value=True,
    yes=False,
):
    msg = prefix + sep.join(i["display"] for i in options) + postfix
    if yes:
        print(msg)
        return yes_value
    else:
        response = raw_input(msg)
        while True:
            if not case_sensitive:
                response = response.lower()
            if strip:
                response = response.strip()
            for expected in options:
                if response in expected["values"]:
                    return expected["value"]
            print(
                'Response "%s" is not one of %s'
                % (
                    response,
                    ", ".join(repr(val) for i in options for val in i["values"]),
                )
            )
            response = raw_input(msg)


def yes_no_prompt(**kwargs):
    return prompt(
        (
            {"value": True, "display": "(y)es", "values": ("y", "yes")},
            {"value": False, "display": "(n)o", "values": ("n", "no")},
        ),
        yes_value=True,
        **kwargs,
    )


# Unfortunately, coqtop -emacs -time reports character locations in
# bytes, as does the glob file, so we need to handle unicode.  Here we
# write a generic slicer based on unicode locations that works across
# versions of python
def slice_string_at_bytes(string, start=None, end=None):
    string = b(string)
    if start is None:
        start = 0
    if end is None:
        end = len(string)
    return s(string[start:end])


def len_in_bytes(string):
    return len(b(string))


def normalize_newlines(string: str) -> str:
    return string.replace("\r\n", "\n").replace("\r", "\n")


# Terminal colors (maybe something cleverer needs to be done for other
# platforms).
class colors:
    ESC = "\033"
    # Escape code doesn't render on github so we use the standard escaped escape.
    # ESC = ''

    HEADER = ESC + "[95m"
    OKBLUE = ESC + "[94m"
    OKCYAN = ESC + "[96m"
    OKGREEN = ESC + "[92m"
    WARNING = ESC + "[93m"
    FAIL = ESC + "[91m"
    ENDC = ESC + "[0m"
    BOLD = ESC + "[1m"
    UNDERLINE = ESC + "[4m"


# Colors a string a given color
# Example usage: color("Hello World!", colors.OKBLUE)
def color(str, col, on=True):
    if not on:
        return str
    else:
        return col + str + colors.ENDC


if sys.version_info < (3, 3):
    import pipes

    shlex_quote = pipes.quote
else:
    import shlex

    shlex_quote = shlex.quote

if sys.version_info < (3, 8):
    shlex_join = lambda split_command: " ".join(shlex_quote(arg) for arg in split_command)
else:
    import shlex

    shlex_join = shlex.join


def resource_path(path):
    if os.path.exists(os.path.join(SCRIPT_DIRECTORY, path)):
        return os.path.join(SCRIPT_DIRECTORY, path)
    if sys.version_info < (3, 7):
        import pkg_resources

        return pkg_resources.resource_filename("coq_tools", path)
    else:
        import importlib.resources

        return importlib.resources.path("coq_tools", path)


def group_by_iter(ls, f):
    """
    Groups elements in a list based on a given condition.

    Args:
        ls (iterable): The list of elements to be grouped.
        f (function): The condition function used for grouping.

    Returns:
        iterable: A list of lists, where each inner list represents a group of elements.
    """
    it = iter(ls)
    prev = next(it)
    cur = [prev]
    for token in it:
        if f(prev, token):
            cur.append(token)
        else:
            yield cur
            cur = [token]
        prev = token
    yield cur


def group_by(ls, f):
    """
    Groups elements in a list based on a given condition.

    Args:
        ls (list): The list of elements to be grouped.
        f (function): The condition function used for grouping.

    Returns:
        list: A list of lists, where each inner list represents a group of elements.

    Example:
        >>> group_by([1, 2, 3, 4, 5, 7, 8, 10], lambda x, y: x + 1 == y)
        [[1, 2, 3, 4, 5], [7, 8], [10]]
    """
    return list(group_by_iter(ls, f))


def list_diff(old_strs: List[str], new_strs: List[str], *, autojunk: bool = False) -> str:
    """
    >>> old = ["line1", "line2", "line3", "line4"]
    >>> new = ["line1", "line3", "line4", "line5"]
    >>> print(list_diff(old, new))
      line1
    - line2
      line3
      line4
    + line5
    """
    # Create a SequenceMatcher object
    sm = SequenceMatcher(None, old_strs, new_strs, autojunk=autojunk)
    opcodes = sm.get_opcodes()
    diff_lines = []

    # Threshold for eliding long equal blocks (if more than this many lines, collapse)
    ELIDE_THRESHOLD = 3

    for tag, i1, i2, j1, j2 in opcodes:
        if tag == "equal":
            block_length = i2 - i1
            if block_length > ELIDE_THRESHOLD:
                # Show the first line, an ellipsis, then the last line of the block.
                diff_lines.append("  " + old_strs[i1])
                diff_lines.append("  ...")
                diff_lines.append("  " + old_strs[i2 - 1])
            else:
                for i in range(i1, i2):
                    diff_lines.append("  " + old_strs[i])
        elif tag == "delete":
            # Lines in old_strs that were removed.
            for i in range(i1, i2):
                diff_lines.append("- " + old_strs[i])
        elif tag == "insert":
            # Lines added in new_strs.
            for j in range(j1, j2):
                diff_lines.append("+ " + new_strs[j])
        elif tag == "replace":
            # Show removed lines
            for i in range(i1, i2):
                diff_lines.append("- " + old_strs[i])
            # Then show inserted lines
            for j in range(j1, j2):
                diff_lines.append("+ " + new_strs[j])

    return "\n".join(diff_lines)


if __name__ == "__main__":
    # if we're working in Python 3.3, we can test this file
    try:
        import doctest

        success = True
    except ImportError:
        print(
            "This is not the main file to use.\nOnly run it if you have doctest (Python 3.3+) and are testing things."
        )
        success = False
    if success:
        doctest.testmod()



# autogenerated by cat /path/to/coq/doc/tools/docgram/fullGrammar | ./generate-full-grammar.py > coq_full_grammar.py
__all__ = ["COQ_GRAMMAR_TOKENS"]

COQ_GRAMMAR_TOKENS = [
    "Abort",
    "About",
    "Add",
    "Admit",
    "Admitted",
    "All",
    "Arguments",
    "As",
    "Assumptions",
    "Axiom",
    "Axioms",
    "Back",
    "BackTo",
    "BinOp",
    "BinOpSpec",
    "BinRel",
    "Bind",
    "Blacklist",
    "Boolean",
    "Canonical",
    "Case",
    "Cd",
    "Check",
    "Class",
    "Classes",
    "Close",
    "CoFixpoint",
    "CoInductive",
    "Coercion",
    "Coercions",
    "Collection",
    "Combined",
    "Comments",
    "Compute",
    "Conjecture",
    "Conjectures",
    "Constant",
    "Constants",
    "Constraint",
    "Constructors",
    "Context",
    "Corollary",
    "Create",
    "CstOp",
    "Cumulative",
    "Custom",
    "Cut",
    "CutOff",
    "Debug",
    "Declare",
    "Defined",
    "Definition",
    "Delimit",
    "Dependencies",
    "Dependency",
    "Dependent",
    "Derive",
    "Diffs",
    "Drop",
    "Elimination",
    "End",
    "Entry",
    "Equality",
    "Equivalent",
    "Eval",
    "Example",
    "Existentials",
    "Existing",
    "Export",
    "Extern",
    "Extra",
    "Extract",
    "Extraction",
    "Fact",
    "Fail",
    "Field",
    "Fields",
    "File",
    "Firstorder",
    "Fixpoint",
    "Flags",
    "Focus",
    "Format",
    "From",
    "Funclass",
    "Function",
    "Functional",
    "GC",
    "Generalizable",
    "Generate",
    "Global",
    "Goal",
    "Grammar",
    "Graph",
    "Guarded",
    "Haskell",
    "Heap",
    "Hint",
    "HintDb",
    "Hints",
    "Hypotheses",
    "Hypothesis",
    "Identity",
    "Immediate",
    "Implicit",
    "Implicits",
    "Import",
    "Include",
    "Induction",
    "Inductive",
    "Infix",
    "Info",
    "Initial",
    "InjTyp",
    "Inline",
    "Inlined",
    "Inspect",
    "Instance",
    "Instances",
    "Intro",
    "Intros",
    "Inversion",
    "Inversion_clear",
    "JSON",
    "Keys",
    "Language",
    "Left",
    "Lemma",
    "Let",
    "Lia",
    "Libraries",
    "Library",
    "Load",
    "LoadPath",
    "Local",
    "Locate",
    "Ltac",
    "Ltac2",
    "ML",
    "Match",
    "Method",
    "Minimality",
    "Mode",
    "Module",
    "Modules",
    "Monomorphic",
    "Morphism",
    "Namespace",
    "Next",
    "No",
    "NoInline",
    "NonCumulative",
    "Notation",
    "Number",
    "OCaml",
    "Obligation",
    "Obligations",
    "Off",
    "On",
    "Opaque",
    "Open",
    "Optimize",
    "Options",
    "Parameter",
    "Parameters",
    "Parametric",
    "Path",
    "Paths",
    "Polymorphic",
    "Prenex",
    "Preterm",
    "Primitive",
    "Print",
    "Private",
    "Profile",
    "Program",
    "Projections",
    "Proof",
    "Prop",
    "PropBinOp",
    "PropOp",
    "PropUOp",
    "Property",
    "Proposition",
    "Pwd",
    "Qed",
    "Quit",
    "Rec",
    "Record",
    "Recursive",
    "Redirect",
    "Reduction",
    "Register",
    "Registered",
    "Relation",
    "Remark",
    "Remove",
    "Require",
    "Reserved",
    "Reset",
    "Resolve",
    "Restart",
    "Rewrite",
    "Right",
    "Ring",
    "Rings",
    "SProp",
    "Saturate",
    "Save",
    "Scheme",
    "Scope",
    "Scopes",
    "Search",
    "SearchPattern",
    "SearchRewrite",
    "Section",
    "Separate",
    "Set",
    "Setoid",
    "Show",
    "Signatures",
    "Solve",
    "Solver",
    "Sort",
    "Sortclass",
    "Sorted",
    "Step",
    "Strategies",
    "Strategy",
    "String",
    "Structure",
    "SubClass",
    "Subgraph",
    "Succeed",
    "SuchThat",
    "Table",
    "Tables",
    "Tactic",
    "Term",
    "Test",
    "TestCompile",
    "Theorem",
    "Time",
    "Timeout",
    "To",
    "Transparent",
    "Type",
    "TypeClasses",
    "Typeclasses",
    "Types",
    "Typing",
    "UnOp",
    "UnOpSpec",
    "Undelimit",
    "Undo",
    "Unfocus",
    "Unfocused",
    "Unfold",
    "Universe",
    "Universes",
    "Unset",
    "Unshelve",
    "Variable",
    "Variables",
    "Variant",
    "Verbose",
    "View",
    "Visibility",
    "Zify",
    "_",
    "abstract",
    "absurd",
    "admit",
    "after",
    "all",
    "and",
    "any",
    "apply",
    "as",
    "assert",
    "assert_succeeds",
    "associativity",
    "assumption",
    "at",
    "auto",
    "autoapply",
    "autorewrite",
    "autounfold",
    "autounfold_one",
    "before",
    "best_effort",
    "beta",
    "bfs",
    "bidirectionality",
    "bigint",
    "binder",
    "bottom",
    "bottomup",
    "btauto",
    "by",
    "case",
    "casetype",
    "cbn",
    "cbv",
    "change",
    "change_no_check",
    "choice",
    "clear",
    "clearbody",
    "closed",
    "cofix",
    "compare",
    "completeness",
    "compute",
    "concl",
    "congr",
    "congruence",
    "constants",
    "constr",
    "constr_eq",
    "constr_eq_nounivs",
    "constr_eq_strict",
    "constructor",
    "context",
    "contradiction",
    "custom",
    "cut",
    "cutoff",
    "cutrewrite",
    "cycle",
    "debug",
    "decidable",
    "decide",
    "decompose",
    "default",
    "delta",
    "dependent",
    "destauto",
    "destruct",
    "dfs",
    "discriminate",
    "discriminated",
    "div",
    "do",
    "eapply",
    "eassert",
    "eassumption",
    "eauto",
    "ecase",
    "econstructor",
    "edestruct",
    "ediscriminate",
    "eelim",
    "eenough",
    "eexact",
    "eexists",
    "einduction",
    "einjection",
    "eintros",
    "eleft",
    "elim",
    "elimtype",
    "else",
    "emp",
    "end",
    "enough",
    "epose",
    "eps",
    "eqn",
    "equality",
    "eremember",
    "erewrite",
    "eright",
    "eset",
    "esimplify_eq",
    "esplit",
    "etransitivity",
    "eval",
    "evar",
    "exact",
    "exact_no_check",
    "exactly_once",
    "exists",
    "expand",
    "external",
    "extra",
    "f_equal",
    "fail",
    "field_lookup",
    "finish_timing",
    "first",
    "firstorder",
    "fix",
    "fold",
    "for",
    "forall",
    "format",
    "fresh",
    "from",
    "fun",
    "functional",
    "gen",
    "generalize",
    "generalize_eqs",
    "generalize_eqs_vars",
    "generally",
    "gfail",
    "gintuition",
    "give_up",
    "global",
    "goal",
    "graph",
    "guard",
    "has_evar",
    "have",
    "head",
    "head_of_constr",
    "headconcl",
    "headhyp",
    "hget_evar",
    "hint",
    "hints",
    "hnf",
    "hresolve_core",
    "hyp",
    "id",
    "ident",
    "idtac",
    "if",
    "implicits",
    "in",
    "induction",
    "infoH",
    "info_auto",
    "info_eauto",
    "info_trivial",
    "injection",
    "inline",
    "innermost",
    "inside",
    "instantiate",
    "into",
    "intro",
    "intros",
    "inversion",
    "inversion_clear",
    "iota",
    "is",
    "is_cofix",
    "is_const",
    "is_constructor",
    "is_evar",
    "is_fix",
    "is_ground",
    "is_ind",
    "is_proj",
    "is_var",
    "isn't",
    "lapply",
    "last",
    "lazy",
    "lazymatch",
    "left",
    "let",
    "level",
    "loss",
    "ltac",
    "ltac1",
    "ltac1val",
    "mapping",
    "match",
    "max",
    "measure",
    "morphism",
    "move",
    "multimatch",
    "mutable",
    "name",
    "native_cast_no_check",
    "native_compute",
    "never",
    "next",
    "no",
    "nomatch",
    "not_evar",
    "notypeclasses",
    "nsatz_compute",
    "numgoals",
    "of",
    "old_hints",
    "once",
    "only",
    "opaque",
    "open_constr",
    "optimize_heap",
    "outermost",
    "outside",
    "par",
    "parsing",
    "pat",
    "pattern",
    "pose",
    "postprocess",
    "power",
    "power_tac",
    "preprocess",
    "printing",
    "profile",
    "profiling",
    "progress",
    "progress_evars",
    "proof",
    "protect_fv",
    "proved",
    "rec",
    "record",
    "red",
    "reference",
    "refine",
    "refl",
    "reflexivity",
    "remember",
    "removed",
    "rename",
    "repeat",
    "replace",
    "reset",
    "restart_timer",
    "return",
    "reverse",
    "revert",
    "revgoals",
    "rewrite",
    "rewrite_db",
    "rewrite_strat",
    "right",
    "ring_lookup",
    "rtauto",
    "scope",
    "scopes",
    "set",
    "setoid",
    "setoid_etransitivity",
    "setoid_reflexivity",
    "setoid_rewrite",
    "setoid_symmetry",
    "setoid_transitivity",
    "shelve",
    "shelve_unifiable",
    "show",
    "sign",
    "signature",
    "simpl",
    "simple",
    "simplify_eq",
    "soft",
    "solve",
    "solve_constraints",
    "specialize",
    "specialize_eqs",
    "split",
    "ssrinstancesofruleL2R",
    "ssrinstancesofruleR2L",
    "ssrinstancesoftpat",
    "start",
    "stepl",
    "stepr",
    "stop",
    "strict",
    "struct",
    "subst",
    "substitute",
    "subterm",
    "subterms",
    "suff",
    "suffices",
    "sum",
    "swap",
    "symmetry",
    "terms",
    "then",
    "time",
    "timeout",
    "top",
    "topdown",
    "transitivity",
    "transparent",
    "transparent_abstract",
    "trivial",
    "try",
    "tryif",
    "type",
    "type_term",
    "typeclasses",
    "under",
    "unfold",
    "unify",
    "unlock",
    "unshelve",
    "until",
    "using",
    "value",
    "via",
    "vm_cast_no_check",
    "vm_compute",
    "warning",
    "wf",
    "where",
    "with",
    "with_strategy",
    "without",
    "wlia",
    "wlog",
    "wlra_Q",
    "wnia",
    "wnra_Q",
    "wpsatz_Q",
    "wpsatz_Z",
    "wsos_Q",
    "wsos_Z",
    "xlia",
    "xlra_Q",
    "xlra_R",
    "xnia",
    "xnra_Q",
    "xnra_R",
    "xpsatz_Q",
    "xpsatz_R",
    "xpsatz_Z",
    "xsos_Q",
    "xsos_R",
    "xsos_Z",
    "zeta",
    "zify_elim_let",
    "zify_iter_let",
    "zify_iter_specs",
    "zify_op",
    "zify_saturate",
]



import re, time
from .Popen_noblock import Popen_async, PIPE, STDOUT, Empty

__all__ = ["join_definitions", "split_statements_to_definitions"]

DEFAULT_VERBOSITY = 1


def DEFAULT_LOG(text, level=DEFAULT_VERBOSITY):
    if level <= DEFAULT_VERBOSITY:
        print(text)


def get_all_nowait_iter(q):
    try:
        while True:
            yield q.get_nowait()
    except Empty:
        pass


def get_all_nowait(q):
    return "".join(get_all_nowait_iter(q))


def get_all_semiwait_iter(q, log=DEFAULT_LOG):
    def log_and_return(val):
        log(val, level=5)
        return val

    try:
        # this is blocking; TODO(jgross): Figure out how to get coqtop
        # to tell us if it's finished computing
        yield log_and_return(q.get(True))
        while True:
            yield log_and_return(q.get(True, 0.1))
    except Empty:
        pass


def get_all_semiwait(q, log=DEFAULT_LOG):
    return "".join(get_all_semiwait_iter(q, log=log))


def get_definitions_diff(previous_definition_string, new_definition_string):
    """Returns a triple of lists (definitions_removed,
    definitions_shared, definitions_added)"""
    old_definitions = [i for i in previous_definition_string.split("|") if i]
    new_definitions = [i for i in new_definition_string.split("|") if i]
    return (
        tuple(i for i in old_definitions if i not in new_definitions),
        tuple(i for i in old_definitions if i in new_definitions),
        tuple(i for i in new_definitions if i not in old_definitions),
    )


# def split_coq_code_to_definitions(code, log=DEFAULT_LOG, coqtop='coqtop'):
#    """Splits coq code into chunks which make up
#    independent definitions/hints/etc, of the form
#
#    {
#     'statements': <list of runnable statements>,
#     'statement':  <entire chunk of code>,
#     'terms_defined': <tuple of terms defined by this chunk of code>,
#     'times': <a list of times for each statement>
#     'output': <a string of the response from coqtop>
#    }"""
#    p = Popen_async([coqtop, '-q', '-emacs', '-time'], stdout=PIPE, stderr=STDOUT, stdin=PIPE)
#    time.sleep(1)


def split_statements_to_definitions(statements, log=DEFAULT_LOG, coqtop="coqtop", coqtop_args=tuple()):
    """Splits a list of statements into chunks which make up
    independent definitions/hints/etc."""
    p = Popen_async([*coqtop, "-q", "-emacs"] + list(coqtop_args), stdout=PIPE, stderr=STDOUT, stdin=PIPE)
    time.sleep(1)
    prompt_reg = re.compile(r"<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*?)</prompt>".replace(" ", r"\s*"))
    defined_reg = re.compile(r"^([^\s]+) is (?:defined|assumed)$", re.MULTILINE)
    # aborted_reg = re.compile(r'^Current goal aborted$', re.MULTILINE)
    # goal_reg = re.compile(r'^\s*=+\s*$', re.MULTILINE)
    # goals and definitions are on stdout, prompts are on stderr
    # clear stdout
    get_all_semiwait(p.stdout, log=log)
    # clear stderr
    # get_all_nowait(p.stderr)

    rtn = []
    cur_definition = {}
    last_definitions = "||"
    cur_definition_names = "||"
    for statement in statements:
        if not statement.strip():
            continue
        log("Write: %s\n\nWait to read..." % statement, level=4)
        p.stdin.write(statement + "\n\n")
        p.stdin.flush()
        stdout = get_all_semiwait(p.stdout, log=log)
        stderr = stdout  # ''.join(get_all_semiwait(p.stderr))

        terms_defined = defined_reg.findall(prompt_reg.sub("", stdout))
        # print((statement, stdout, terms_defined))
        prompt_match = prompt_reg.search(stderr)

        if not prompt_match or len(prompt_match.groups()) != 5:
            if not prompt_match:
                log("Likely fatal warning: I did not recognize the output from coqtop:")
                log("stdout: %s\nstderr: %s" % (repr(stdout), repr(stderr)))
                log(
                    "I will append the current statement (%s) to the list of definitions as-is, but I don't expect this to work."
                    % statement
                )
            else:
                log("Crazy things are happening; the number of groups isn't what it should be (should be 5 groups):")
                log(
                    "prompt_match.groups(): %s\nstdout: %s\nstderr: %s\nstatement: %s\n"
                    % (repr(prompt_match.groups()), repr(stdout), repr(stderr), repr(statement))
                )

            log((statement, terms_defined, cur_definition_names, cur_definition.get(cur_definition_names, [])), level=2)
            if cur_definition_names.strip("|"):
                cur_definition[cur_definition_names]["statements"].append(statement)
                cur_definition[cur_definition_names]["terms_defined"] += terms_defined
            else:
                rtn.append({"statements": (statement,), "statement": statement, "terms_defined": tuple(terms_defined)})
        else:
            cur_name, line_num1, cur_definition_names, line_num2, unknown = prompt_reg.search(stderr).groups()
            definitions_removed, definitions_shared, definitions_added = get_definitions_diff(
                last_definitions, cur_definition_names
            )

            # first, to be on the safe side, we add the new
            # definitions key to the dict, if it wasn't already there.
            if cur_definition_names.strip("|") and cur_definition_names not in cur_definition:
                cur_definition[cur_definition_names] = {"statements": [], "terms_defined": []}

            log(
                (
                    statement,
                    terms_defined,
                    last_definitions,
                    cur_definition_names,
                    cur_definition.get(last_definitions, []),
                    cur_definition.get(cur_definition_names, []),
                ),
                level=2,
            )

            # first, we handle the case where we have just finished
            # defining something.  This should correspond to
            # len(definitions_removed) > 0 and len(terms_defined) > 0.
            # If only len(definitions_removed) > 0, then we have
            # aborted something.  If only len(terms_defined) > 0, then
            # we have defined something with a one-liner.
            if definitions_removed:
                cur_definition[last_definitions]["statements"].append(statement)
                cur_definition[last_definitions]["terms_defined"] += terms_defined
                if cur_definition_names.strip("|"):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]["statements"] += cur_definition[last_definitions]["statements"]
                    cur_definition[cur_definition_names]["terms_defined"] += cur_definition[last_definitions][
                        "terms_defined"
                    ]
                    del cur_definition[last_definitions]
                else:
                    # we're at top-level, so add this as a new
                    # definition
                    rtn.append(
                        {
                            "statements": tuple(cur_definition[last_definitions]["statements"]),
                            "statement": "\n".join(cur_definition[last_definitions]["statements"]),
                            "terms_defined": tuple(cur_definition[last_definitions]["terms_defined"]),
                        }
                    )
                    del cur_definition[last_definitions]
                    # print('Adding:')
                    # print(rtn[-1])
            elif terms_defined:
                if cur_definition_names.strip("|"):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]["statements"].append(statement)
                    cur_definition[cur_definition_names]["terms_defined"] += terms_defined
                else:
                    # we're at top level, so add this as a new
                    # definition
                    rtn.append(
                        {"statements": (statement,), "statement": statement, "terms_defined": tuple(terms_defined)}
                    )

            # now we handle the case where we have just opened a fresh
            # definition.  We've already added the key to the
            # dictionary.
            elif definitions_added:
                # print(definitions_added)
                cur_definition[cur_definition_names]["statements"].append(statement)
            else:
                # if we're in a definition, append the statement to
                # the queue, otherwise, just add it as it's own
                # statement
                if cur_definition_names.strip("|"):
                    cur_definition[cur_definition_names]["statements"].append(statement)
                else:
                    rtn.append({"statements": (statement,), "statement": statement, "terms_defined": tuple()})
        last_definitions = cur_definition_names

    log((last_definitions, cur_definition_names), level=2)
    if last_definitions.strip("||"):
        rtn.append(
            {
                "statements": tuple(cur_definition[cur_definition_names]["statements"]),
                "statement": "\n".join(cur_definition[cur_definition_names]["statements"]),
                "terms_defined": tuple(cur_definition[cur_definition_names]["terms_defined"]),
            }
        )
        del cur_definition[last_definitions]
    # for i in rtn:
    # print(i)
    p.stdin.close()
    return rtn


def join_definitions(definitions):
    return "\n".join(i["statement"] for i in definitions)



from .util import group_by

__all__ = ["strip_comments", "strip_trailing_comments"]


def annotate_with_is_comment(tokens):
    """
    Annotates each token in the given list with a boolean value indicating whether it is a comment or not.

    Args:
        tokens (list[str]): The list of tokens to be annotated.

    Yields:
        Tuple[bool, str]: A tuple containing a boolean value indicating whether
        the token is a comment or not, and the token itself.
    """
    is_string = False
    comment_level = 0
    for token in tokens:
        is_comment = comment_level > 0
        if is_string:
            if token.count('"') % 2 == 1:
                # there are an odd number of '"' characters, indicating that we've ended the string
                is_string = False
        elif token.count('"') % 2 == 1:
            # there are an odd number of '"' characters, so we're starting a string
            is_string = True
        elif token == "(*":
            comment_level += 1
            is_comment = True
        elif comment_level > 0 and token == "*)":
            comment_level -= 1
        yield is_comment, token


def split_comments(contents: str):
    """
    Splits the given string into a list of blocks, each of which is either a comment or is entirely comment-free
    """
    contents = contents.replace("(*", " (* ").replace("*)", " *) ")
    tokens = contents.split(" ")
    token_groups = group_by(annotate_with_is_comment(tokens), lambda x, y: x[0] == y[0])
    rtn = []
    for group in token_groups:
        group = " ".join(token for _, token in group)
        if group.startswith("(*"):
            group = " " + group
        if group.endswith("*)"):
            group = group + " "
        group = group.replace(" (* ", "(*").replace(" *) ", "*)")
        rtn.append(group)
    return rtn


def strip_comments(contents):
    """Strips the comments from coq code in contents.

    The strings in contents are only preserved if there are no
    comment-like tokens inside of strings.  Stripping should be
    successful and correct, regardless of whether or not there are
    comment-like tokens in strings.

    The behavior of this method is undefined if there are any
    notations which change the meaning of '(*', '*)', or '"'.

    Note that we take some extra care to leave *) untouched when it
    does not terminate a comment.
    """
    return "".join(block for block in split_comments(contents) if not (block.startswith("(*") and block.endswith("*)")))


def strip_trailing_comments(contents, strip_chars=None):
    """Strips the trailing comments from coq code in contents.

    The definition of "trailing" is that all non-comment characters after the
    comment must be removed with str.strip(strip_chars).
    """
    blocks = split_comments(contents)
    trailing = []
    for block_i, block in reversed(list(enumerate(blocks))):
        if not block.strip(strip_chars):
            trailing.insert(0, block)
        if block.startswith("(*") and block.endswith("*)"):
            continue
        return "".join(blocks[: block_i + 1] + trailing)
    return "".join(trailing)



#!/usr/bin/env python
import os, os.path
from .argparse_compat import argparse
from .import_util import IMPORT_ABSOLUTIZE_TUPLE, ALL_ABSOLUTIZE_TUPLE
from .custom_arguments import (
    add_libname_arguments,
    update_env_with_libnames,
    update_env_with_coqpath_folders,
    add_logging_arguments,
    process_logging_arguments,
    get_parser_name_mapping,
)
from .coq_version import get_coqc_coqlib
from .replace_imports import include_imports

__all__ = ["main"]

parser = argparse.ArgumentParser(description="Inline the imports of a file")
parser.add_argument(
    "input_file",
    metavar="IN_FILE",
    type=argparse.FileType("r"),
    help="a .v file to inline the imports of",
)
parser.add_argument(
    "output_file",
    metavar="OUT_FILE",
    type=argparse.FileType("w"),
    help="a .v file to write to",
)
parser.add_argument(
    "--fast-merge-imports",
    dest="fast_merge_imports",
    action="store_const",
    const=True,
    default=False,
    help="Use a faster method for combining imports",
)
parser.add_argument(
    "--no-deps",
    dest="use_coq_makefile_for_deps",
    action="store_const",
    const=False,
    default=True,
    help=("Don't do dependency analysis with coq_makefile."),
)
parser.add_argument(
    "--no-pwd-deps",
    dest="walk_tree",
    action="store_const",
    const=False,
    default=True,
    help=("Don't add all files in the current directory to the dependency analysis."),
)
parser.add_argument(
    "--no-wrap-modules",
    dest="wrap_modules",
    action="store_const",
    const=False,
    default=True,
    help=(
        "Don't wrap imports in Modules.  By default, the "
        + "contents of each file is wrapped in its own "
        + "module to deal with renaming issues.  This "
        + "can cause issues with subdirectories."
    ),
)
parser.add_argument(
    "--absolutize-constants",
    dest="absolutize",
    action="store_const",
    default=IMPORT_ABSOLUTIZE_TUPLE,
    const=ALL_ABSOLUTIZE_TUPLE,
    help=(
        "Replace constants with fully qualified versions.  "
        + "By default, all constants are not fully qualified.  If you have "
        + "many overlapping file names in different directories "
        + "and use partially qualified names that differ depending "
        + "on which files have been Required, not absolutizing constants "
        + "may cause name resolution to fail."
    ),
)
parser.add_argument(
    "--coqbin",
    metavar="COQBIN",
    dest="coqbin",
    type=str,
    default="",
    help="The path to a folder containing the coqc and coqtop programs.",
)
parser.add_argument(
    "--coqc",
    metavar="COQC",
    dest="coqc",
    type=str,
    default=None,
    action="append",
    help="The path to the coqc program.",
)
# parser.add_argument(
#     "--coqtop",
#     metavar="COQTOP",
#     dest="coqtop",
#     type=str,
#     default=None,
#     action="append",
#     help=("The path to the coqtop program (default: %s)." % DEFAULT_COQTOP),
# )
parser.add_argument(
    "--coqc-args",
    metavar="ARG",
    dest="coqc_args",
    type=str,
    nargs="?",
    help='Arguments to pass to coqc; e.g., " -indices-matter" (leading and trailing spaces are stripped)',
)
# parser.add_argument(
#     "--coqtop-args",
#     metavar="ARG",
#     dest="coqtop_args",
#     type=str,
#     nargs="?",
#     help='Arguments to pass to coqtop; e.g., " -indices-matter" (leading and trailing spaces are stripped)',
# )
parser.add_argument(
    "--coq_makefile",
    metavar="COQ_MAKEFILE",
    dest="coq_makefile",
    type=str,
    default="coq_makefile",
    action="append",
    help="The path to the coq_makefile program.",
)
parser.add_argument(
    "--inline-user-contrib",
    dest="inline_user_contrib",
    action="store_const",
    const=True,
    default=False,
    help=("Attempt to inline requires from the user-contrib folder"),
)
parser.add_argument(
    "--no-inline-stdlib",
    dest="no_inline_stdlib",
    action="store_const",
    const=True,
    default=False,
    help=("Skip Stdlib directory when using --inline-user-contrib"),
)
parser.add_argument(
    "--no-inline-corelib",
    dest="no_inline_corelib",
    action="store_const",
    const=True,
    default=False,
    help=("Skip Corelib directory when using --inline-user-contrib"),
)
add_libname_arguments(parser)
add_logging_arguments(parser)


def main():
    args = process_logging_arguments(parser.parse_args())

    def prepend_coqbin(prog):
        if isinstance(prog, str):
            prog = (prog,)
        if args.coqbin != "":
            return (os.path.join(args.coqbin, prog[0]), *prog[1:])
        else:
            return tuple(prog)

    env = {
        "log": args.log,
        "coqc": prepend_coqbin(args.coqc or "coqc"),
        "absolutize": args.absolutize,
        "as_modules": args.wrap_modules,
        "fast": args.fast_merge_imports,
        "coq_makefile": prepend_coqbin(args.coq_makefile or "coq_makefile"),
        "use_coq_makefile_for_deps": args.use_coq_makefile_for_deps,
        "walk_tree": args.walk_tree,
        "cli_mapping": get_parser_name_mapping(parser),
    }
    update_env_with_libnames(env, args)
    if args.inline_user_contrib:
        # Build list of directories to skip based on command-line arguments
        skip_dirs = []
        if args.no_inline_stdlib:
            skip_dirs.append("Stdlib")
        if args.no_inline_corelib:
            skip_dirs.append("Corelib")
        update_env_with_coqpath_folders(
            "",
            env,
            os.path.join(
                get_coqc_coqlib(env["coqc"], coq_args=env["coqc_args"], **env),
                "user-contrib",
            ),
            skip_dirs=skip_dirs,
        )

    filename = args.input_file.name
    args.input_file.close()

    rtn = include_imports(filename, **env)

    args.output_file.write(rtn)
    args.output_file.close()


if __name__ == "__main__":
    main()
